"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.analyzeReactComponent = void 0;
const type_analyzer_1 = require("@previewjs/type-analyzer");
const typescript_1 = __importDefault(require("typescript"));
function analyzeReactComponent(typeAnalyzer, component, args, propTypes) {
    let resolved = computePropsType(typeAnalyzer, component.signature, propTypes);
    let providedArgs = new Set();
    if (args) {
        const argsType = typeAnalyzer.checker.getTypeAtLocation(args);
        providedArgs = new Set(argsType.getProperties().map((prop) => prop.name));
    }
    return {
        name: component.name,
        propsType: resolved.type,
        types: { ...resolved.collected },
        providedArgs,
    };
}
exports.analyzeReactComponent = analyzeReactComponent;
function computePropsType(typeAnalyzer, signature, propTypes) {
    if (propTypes) {
        return computePropsTypeFromPropTypes(typeAnalyzer, propTypes);
    }
    return computePropsTypeFromSignature(typeAnalyzer, signature);
}
function computePropsTypeFromSignature(typeAnalyzer, signature) {
    const firstParam = signature.getParameters()[0];
    if (!firstParam) {
        return {
            type: type_analyzer_1.EMPTY_OBJECT_TYPE,
            collected: {},
        };
    }
    if (!firstParam.valueDeclaration) {
        return {
            type: type_analyzer_1.UNKNOWN_TYPE,
            collected: {},
        };
    }
    const type = typeAnalyzer.checker.getTypeOfSymbolAtLocation(firstParam, firstParam.valueDeclaration);
    try {
        let { type: propsType, collected } = typeAnalyzer.resolveType(type);
        [propsType] = (0, type_analyzer_1.resolveType)(propsType, collected, []);
        stripUnusedProps: if (propsType.kind === "object" &&
            typescript_1.default.isParameter(firstParam.valueDeclaration)) {
            if (typescript_1.default.isObjectBindingPattern(firstParam.valueDeclaration.name)) {
                const bindingPattern = firstParam.valueDeclaration.name;
                const usedProps = new Set();
                for (const element of bindingPattern.elements) {
                    if (element.dotDotDotToken) {
                        break stripUnusedProps;
                    }
                    const elementName = element.propertyName || element.name;
                    if (!typescript_1.default.isIdentifier(elementName)) {
                        break stripUnusedProps;
                    }
                    usedProps.add(elementName.text);
                }
                propsType = (0, type_analyzer_1.objectType)(Object.fromEntries(Object.entries(propsType.fields).filter(([key]) => usedProps.has(key))));
            }
        }
        return { type: propsType, collected };
    }
    catch (e) {
        console.warn(`Unable to resolve props type for ${typeAnalyzer.checker.typeToString(type)}`, e);
    }
    return {
        type: type_analyzer_1.UNKNOWN_TYPE,
        collected: {},
    };
}
function computePropsTypeFromPropTypes(typeAnalyzer, propTypes) {
    const type = typeAnalyzer.checker.getTypeAtLocation(propTypes);
    const fields = {};
    let collected = {};
    for (const property of type.getProperties()) {
        fields[property.name] = (() => {
            if (!property.valueDeclaration) {
                return type_analyzer_1.UNKNOWN_TYPE;
            }
            const propertyType = typeAnalyzer.checker.getTypeAtLocation(property.valueDeclaration);
            const typeArguments = typeAnalyzer.resolveTypeArguments(propertyType);
            const fieldType = typeArguments.types[0];
            if (fieldType) {
                collected = {
                    ...collected,
                    ...typeArguments.collected,
                };
                return (0, type_analyzer_1.maybeOptionalType)(fieldType, propertyType.symbol.name === "Requireable");
            }
            return type_analyzer_1.UNKNOWN_TYPE;
        })();
    }
    return {
        type: (0, type_analyzer_1.objectType)(fields),
        collected: {},
    };
}
