"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.extractReactComponents = void 0;
const ts_helpers_1 = require("@previewjs/core/ts-helpers");
const typescript_1 = __importDefault(require("typescript"));
function extractReactComponents(program, filePath) {
    var _a, _b, _c;
    const sourceFile = program.getSourceFile(filePath);
    if (!sourceFile) {
        return [];
    }
    const checker = program.getTypeChecker();
    let components = [];
    const nameToExportedName = (0, ts_helpers_1.detectExportedNames)(sourceFile);
    for (const statement of sourceFile.statements) {
        if (typescript_1.default.isExportAssignment(statement)) {
            const signature = extractReactComponent(checker, statement.expression);
            if (signature) {
                components.push({
                    filePath,
                    name: "default",
                    exported: true,
                    offsets: [[statement.getStart(), statement.getEnd()]],
                    signature,
                });
            }
        }
        else if (typescript_1.default.isVariableStatement(statement)) {
            for (const declaration of statement.declarationList.declarations) {
                if (!typescript_1.default.isIdentifier(declaration.name) || !declaration.initializer) {
                    continue;
                }
                const name = declaration.name.text;
                const exportedName = nameToExportedName[name];
                if (!isValidReactComponentName(name)) {
                    continue;
                }
                const signature = extractReactComponent(checker, declaration.initializer);
                if (signature) {
                    components.push({
                        filePath,
                        name,
                        exported: !!exportedName,
                        offsets: [[statement.getStart(), statement.getEnd()]],
                        signature,
                    });
                }
            }
        }
        else if (typescript_1.default.isFunctionDeclaration(statement)) {
            const isDefaultExport = !!((_a = statement.modifiers) === null || _a === void 0 ? void 0 : _a.find((m) => m.kind === typescript_1.default.SyntaxKind.ExportKeyword)) &&
                !!((_b = statement.modifiers) === null || _b === void 0 ? void 0 : _b.find((m) => m.kind === typescript_1.default.SyntaxKind.DefaultKeyword));
            const name = (_c = statement.name) === null || _c === void 0 ? void 0 : _c.text;
            const exported = (name && !!nameToExportedName[name]) || isDefaultExport;
            if (isDefaultExport || (name && isValidReactComponentName(name))) {
                const signature = extractReactComponent(checker, statement);
                if (signature) {
                    components.push({
                        filePath,
                        name: name || "default",
                        exported,
                        offsets: [[statement.getStart(), statement.getEnd()]],
                        signature,
                    });
                }
            }
        }
        else if (typescript_1.default.isClassDeclaration(statement) && statement.name) {
            const name = statement.name.text;
            const exportedName = nameToExportedName[name];
            if (!isValidReactComponentName(name)) {
                continue;
            }
            const signature = extractReactComponent(checker, statement);
            if (signature) {
                components.push({
                    filePath,
                    name,
                    exported: !!exportedName,
                    offsets: [[statement.getStart(), statement.getEnd()]],
                    signature,
                });
            }
        }
    }
    return components;
}
exports.extractReactComponents = extractReactComponents;
function extractReactComponent(checker, node) {
    const type = checker.getTypeAtLocation(node);
    // Function component.
    for (const callSignature of type.getCallSignatures()) {
        if (isValidComponentReturnType(callSignature.getReturnType())) {
            return callSignature;
        }
    }
    // Class component.
    if (type.symbol) {
        const classType = checker.getTypeOfSymbolAtLocation(type.symbol, node);
        for (const constructSignature of classType.getConstructSignatures()) {
            const returnType = constructSignature.getReturnType();
            if (returnType.getProperty("render")) {
                return constructSignature;
            }
        }
    }
    return null;
}
function isValidComponentReturnType(type) {
    if (isJsxElement(type)) {
        return true;
    }
    return false;
}
const jsxElementTypes = new Set(["Element", "ReactElement"]);
function isJsxElement(type) {
    var _a;
    if (type.isUnion()) {
        for (const subtype of type.types) {
            if (isJsxElement(subtype)) {
                return true;
            }
        }
    }
    return jsxElementTypes.has((_a = type.symbol) === null || _a === void 0 ? void 0 : _a.getEscapedName().toString());
}
function isValidReactComponentName(name) {
    return name.length > 0 && name[0] >= "A" && name[0] <= "Z";
}
