"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.reactFrameworkPlugin = void 0;
const vfs_1 = require("@previewjs/core/vfs");
const path_1 = __importDefault(require("path"));
const typescript_1 = __importDefault(require("typescript"));
const analyze_component_1 = require("./analyze-component");
const args_1 = require("./args");
const component_loader_plugin_1 = require("./component-loader-plugin");
const extract_component_1 = require("./extract-component");
const optimize_deps_plugin_1 = require("./optimize-deps-plugin");
const prop_types_1 = require("./prop-types");
const react_imports_plugin_1 = require("./react-imports-plugin");
const special_types_1 = require("./special-types");
const svgr_plugin_1 = require("./svgr-plugin");
exports.reactFrameworkPlugin = {
    isCompatible: async (dependencies) => {
        const react = dependencies["react"];
        if (!react) {
            return false;
        }
        return react.majorVersion >= 16;
    },
    async create({ svgr } = {}) {
        const previewDirPath = path_1.default.resolve(__dirname, "..", "preview");
        return {
            name: "@previewjs/plugin-react",
            defaultWrapperPath: "__previewjs__/Wrapper.tsx",
            previewDirPath,
            tsCompilerOptions: {
                jsx: typescript_1.default.JsxEmit.ReactJSX,
                jsxImportSource: "react",
            },
            transformReader: (reader, rootDirPath) => (0, vfs_1.createStackedReader)([
                reader,
                (0, vfs_1.createFileSystemReader)({
                    mapping: {
                        from: path_1.default.join(previewDirPath, "types"),
                        to: path_1.default.join(rootDirPath, "node_modules", "@types"),
                    },
                    watch: false,
                }),
            ]),
            componentDetector: (program, filePaths) => {
                const components = [];
                for (const filePath of filePaths) {
                    components.push(...(0, extract_component_1.extractReactComponents)(program, filePath));
                }
                return components;
            },
            componentAnalyzer: ({ typescriptAnalyzer, getTypeAnalyzer }) => (filePath, componentName) => {
                const program = typescriptAnalyzer.analyze([filePath]);
                const typeAnalyzer = getTypeAnalyzer(program, special_types_1.REACT_SPECIAL_TYPES);
                const component = (0, extract_component_1.extractReactComponents)(program, filePath).find((c) => c.name === componentName);
                if (!component) {
                    throw new Error(`Component ${componentName} was not found in ${filePath}`);
                }
                const sourceFile = program.getSourceFile(filePath);
                let args = null;
                let propTypes = null;
                if (sourceFile) {
                    args = (0, args_1.detectArgs)(sourceFile, component.name);
                    propTypes = (0, prop_types_1.detectPropTypes)(sourceFile, component.name);
                }
                return (0, analyze_component_1.analyzeReactComponent)(typeAnalyzer, component, args, propTypes);
            },
            viteConfig: (config) => {
                return {
                    plugins: [
                        (0, optimize_deps_plugin_1.optimizeReactDepsPlugin)(),
                        (0, component_loader_plugin_1.reactComponentLoaderPlugin)({
                            config,
                        }),
                        (0, svgr_plugin_1.svgrPlugin)({
                            exportedComponentName: (svgr === null || svgr === void 0 ? void 0 : svgr.componentName) || "ReactComponent",
                            alias: config.alias,
                        }),
                        (0, react_imports_plugin_1.reactImportsPlugin)(),
                    ],
                    define: {
                        "process.env.RUNNING_INSIDE_PREVIEWJS": "1",
                    },
                };
            },
        };
    },
};
